/************************************************************************
 * 
 * This sound engine is based on:
 * The UnrealSpeccy emulator by SMT.
 *
 * Thanks.
 * 
 ************************************************************************
 */

#define DIRECTSOUND_VERSION 0x0600
#include <dsound.h>


class   ay_emulator 
{
public:

inline
ay_emulator()
{
//    zero_memory(this, sizeof(synthesizer));
    for( int j = 2 * AY_channels - 1; j >= 0; j-- )
        for( int i = 16; i; )
            volume[j][--i] = ( volume_table[i] * ZX_AY_volume >> 16 )
                           * stereo[j] >> 8;
    noise_state |= 0xFFFF;
}

void inline kill()
{
    stop();
    if( sound_buffer)
        sound_buffer->Release();
    sound_buffer = 0;
    if( direct_sound )
        direct_sound->Release();
    direct_sound = 0;
}

static unsigned long __stdcall
play(void * p);

int inline
initialize(HWND window_handle)   
{
    WAVEFORMATEX    wave_format;
    DSBUFFERDESC    soundbuffer_descriptor;

    //zero_memory(&wave_format, sizeof(wave_format));
    //zero_memory(&soundbuffer_descriptor, sizeof(soundbuffer_descriptor));
    soundbuffer_descriptor.dwSize  |= sizeof(DSBUFFERDESC);
    soundbuffer_descriptor.dwFlags |= DSBCAPS_GETCURRENTPOSITION2
                                   |  DSBCAPS_GLOBALFOCUS;
    soundbuffer_descriptor.dwBufferBytes |= soundbuffer_size;
    soundbuffer_descriptor.lpwfxFormat = &wave_format;

    enum {  samples_per_sec = soundcard_frequency,
            channels        = 2,    bits_per_sample = 16,
            block_align     = bits_per_sample / 8 * channels
    };    
    wave_format.wFormatTag      |= WAVE_FORMAT_PCM;
    wave_format.nSamplesPerSec  |= samples_per_sec;
    wave_format.nChannels       |= channels;
    wave_format.wBitsPerSample  |= bits_per_sample;
    wave_format.nBlockAlign     |= block_align;
    wave_format.nAvgBytesPerSec |= samples_per_sec * block_align;
    
  //  soundbuffer_position = 0;
    PRINTF("initializing DirectSound...\r\n");
    int(__stdcall * DirectSoundCreate)(GUID *, IDirectSound **, void*)
        = (int(__stdcall *)(GUID*, IDirectSound **, void*))
        GetProcAddress(LoadLibraryA("dsound"),"DirectSoundCreate");
    if( ! DirectSoundCreate )
        return panic("DirectX 7 or later not installed");
    else if( DirectSoundCreate(0, &direct_sound, 0) )
        return panic("couldn't initialize DirectSound");
    else if( direct_sound->SetCooperativeLevel(window_handle, DSSCL_EXCLUSIVE) )
        return panic("couldn't set proper cooperative level");
    else if( direct_sound->CreateSoundBuffer(&soundbuffer_descriptor, 
                        &sound_buffer, 0) )
        return panic("couldn't create sound buffer");
    PRINTF("Ok.\r\n");
    return 0;
}


protected:

    IDirectSound          * direct_sound;
    IDirectSoundBuffer    * sound_buffer;

    enum 
    {
        tick_FF                 = 6,
        tick_F                  = 1 << tick_FF,
        ZX_frame_frequency      = 50,
        ZX_ticks_per_frame      = 69888,
        ZX_AY_volume            = 5000,
        ZX_AY_frequency         = 1773400,

        soundcard_frequency     = 44100,
        samples_per_frame       = soundcard_frequency / ZX_frame_frequency,
        soundbuffer_chunk_size  = samples_per_frame * 2*sizeof(short),
        sount_ticks_per_frame   = samples_per_frame << tick_FF,

        soundbuffer_size        = 2 * soundcard_frequency * 2*sizeof(short) // sec
    };
    unsigned                soundbuffer_position;

    enum {  channel_A,  channel_B,  channel_C,  AY_channels };

    static unsigned long    local_buffer    [2*samples_per_frame];
    static unsigned short   volume_table    [16];
    static unsigned short   stereo          [2*AY_channels];
    static unsigned         volume          [2*AY_channels][16];

    unsigned                left, right;
    unsigned                tone_cycle      [AY_channels];
    unsigned                tone_frequency  [AY_channels];
    unsigned                tone_mask       [AY_channels];
    unsigned                tone_bit        [AY_channels];
    unsigned                tone_volume     [AY_channels];
    unsigned                noise_cycle,    noise_frequency;
    unsigned                noise_mask      [AY_channels];
    unsigned                noise_bit,      noise_state;
    unsigned                envelope_cycle, envelope_frequency;
    unsigned                envelope_volume;
    int                     envelope_add;
    unsigned char           envelope_mode;
    

void clear_buffer()
{
    void          * start;
    unsigned long   size;
    if( sound_buffer &&
      ! sound_buffer->Lock(0, 0, &start, &size, 0, 0, DSBLOCK_ENTIREBUFFER) )
    {
        zero_memory(start, size);
        sound_buffer->Unlock(start, size, 0, 0);
    }
}

void stop()
{
    if(sound_buffer)
    {
        sound_buffer->Stop();
        clear_buffer();
    }
}


void inline
do_sound()
{
    while( true )
    {
        unsigned long   write_position;
        sound_buffer->GetCurrentPosition(0, &write_position);
        if( soundbuffer_position <= write_position )
            break;
        sound_buffer->Play(0, 0, DSBPLAY_LOOPING);
        Sleep(1);
    }
    void              * write_cursor;
    unsigned long       size;
    if( sound_buffer->Lock(soundbuffer_position, soundbuffer_chunk_size,
                            &write_cursor, &size, 0, 0, 0) )
        sound_buffer->Restore();
    else
    {   //   +  = 2 + 2 .
        copy_by_dwords(write_cursor, &local_buffer, size);
        sound_buffer->Unlock(write_cursor, size, 0, 0);
        soundbuffer_position += soundbuffer_chunk_size;
        if( soundbuffer_position >= soundbuffer_size )
            soundbuffer_position = 0;
    }
}


void flush_digital_sound(unsigned sndtick, unsigned endtick)
{
    unsigned    scale,  p = sndtick >> tick_FF;
    if( (sndtick ^ endtick) & ~(tick_F-1) ) 
    {
        scale = tick_F - (sndtick & (tick_F-1));
        local_buffer[p++] += ((left*scale) >> tick_FF)
                          + (((right*scale) >> tick_FF) << 16);
        sndtick = (sndtick | (tick_F-1)) + 1;
        if( (sndtick ^ endtick) & ~(tick_F-1) )
            do 
            {
                local_buffer[p++] += left + (right << 16);
                sndtick += tick_F;
            }while( (sndtick ^ endtick) & ~(tick_F-1) );
    }
    scale = endtick - sndtick;
    local_buffer[p++] += ((left*scale) >> tick_FF)
                      + (((right*scale) >> tick_FF) << 16);
}


void inline
flush_sound()
{
    zero_memory(local_buffer, soundbuffer_chunk_size);

    unsigned tick = 0;  unsigned t = 0;
    while (t < ZX_AY_frequency / ZX_frame_frequency >> 3)
    {
        t++;
        if( ++noise_cycle >= noise_frequency )
        {
            noise_cycle = 0;
            noise_state = (noise_state * 2 + 1) ^ 
                ( ((noise_state >> 16)^(noise_state >> 13)) & 1 );
            noise_bit = - ((noise_state >> 16) & 1);
        }
        if( ++envelope_cycle >= envelope_frequency )
        {
            envelope_cycle = 0;
            envelope_volume += envelope_add;
            if( envelope_volume & ~31 )
            {
                unsigned mask = 1 << envelope_mode;
                if( mask & ( 1<<0 | 1<<1 | 1<<2 | 1<<3 | 1<<4 |
                             1<<5 | 1<<6 | 1<<7 | 1<<9 | 1<<15 ) )
                    envelope_volume = envelope_add = 0;
                else if( mask & ( 1<<8 | 1<<12 ) )
                    envelope_volume &= 31;
                else if( mask & ( 1<<10 | 1<<14 ) )
                    envelope_add = -envelope_add, envelope_volume += envelope_add;
                else
                    envelope_volume = 31,   envelope_add = 0; //11,13
            }
        }
        left =  right= 0;
        for(int channel = AY_channels; channel; )
        {
            if( ++tone_cycle[--channel] >= tone_frequency[channel] )
            {
                tone_cycle[channel] = 0;
                tone_bit[channel] ^= -1;
            }
            unsigned en, bit;
            en = tone_volume[channel] & 0x10 ? envelope_volume / 2 
                                             : tone_volume[channel] & 0x0F;
            bit = (tone_bit[channel] | tone_mask[channel])
                &         (noise_bit | noise_mask[channel]);
            left  += volume[channel*2+0][en] & bit;
            right += volume[channel*2+1][en] & bit;
        }
        unsigned neartick = t * ( (1 << tick_FF) / 2
                + (unsigned)((__int64)(soundcard_frequency << tick_FF)
                * 8 * 16384 / ZX_AY_frequency) ) / 16384; 
        flush_digital_sound(tick, neartick);
        tick = neartick;
    }
    flush_digital_sound(tick, sount_ticks_per_frame);
}

};


unsigned short  ay_emulator::volume_table[] =
{   
    0x0000,0x0340,0x04C0,0x06F2,0x0A44,0x0F13,0x1510,0x227E,
    0x289F,0x414E,0x5B21,0x7258,0x905E,0xB550,0xD7A0,0xFFFF
};

unsigned short  ay_emulator::stereo[] = 
    #define PERCENT(X)(0x100*X/100)
{   // left         // right        // channels
    PERCENT(100),   PERCENT(10),    // A
    PERCENT(66),    PERCENT(66),    // B
    PERCENT(10),    PERCENT(100)    // C
    #undef PERCENT
};

unsigned        ay_emulator::volume[2*AY_channels][16];

unsigned long   ay_emulator::local_buffer[];

// EOF
