/************************************************************************
 * 
 ************************************************************************
 */ 


const   H_tiles = 30;
const   V_tiles = 20;

const   H_tile_size = 21;
const   V_tile_size = 20;

//=========================================================================//
//    
// ""    -     #3 
unsigned char   room[V_tiles+1][H_tiles];

enum    tiles
{ 
        empty,      // 0
        block,      // 1
        brick,      // 2
        bridge,     // 3
        dummy,      // 4      = 20
        runner,     // 5     , .
        treasure,   // 6
        ladder,     // 7
        under_bridge,

        bits_per_tile = 3,
        tiles_in_block = 8*sizeof(short)/bits_per_tile
};

void inline cache_room_data(
        short (* level_data)[V_tiles * H_tiles / tiles_in_block] 
        )
{
    static  short * cached_level_data;
    short * data = (short *)level_data;
    if( cached_level_data != data )
        //    x  y
        for( unsigned i = 0; i < V_tiles * H_tiles; i++ )
        {
            unsigned    block;
            if( ! (i % tiles_in_block) )
                block = *data++;
            room[1][i] = block>>(bits_per_tile*(tiles_in_block-1)) 
                       & (1<<bits_per_tile) - 1;

            switch( room[1][i] )
            {
            case dummy:
                room[1][i] = empty;
                goto check_under_bridge;

            case runner:
                room[1][i] = empty;

            case empty:
            case treasure:
            check_under_bridge:
                if( ! (room[0][i] & under_bridge) ) //  
                    break;
            //      " "
            case bridge:
                room[1][i] |= under_bridge;
                break;
            }
            block <<= bits_per_tile;        
        }
}


//=========================================================================//
// # 0
void draw_tile_empty(ARGB * memory, long pitch, unsigned xpos, unsigned ypos)
{
    ypos = ypos * V_tile_size - (Frame_step>>16);
    xpos = xpos * H_tile_size + (Frame_step>>17);
    for( unsigned y = V_tile_size; y; y-- )
    {
        ARGB c = 0x3f0000 & qsin(qsin(ypos)>>24)>>10;
        for( unsigned x = 0; x < H_tile_size; x++)
            memory[x] = c | 0x3f &(qsin(qsin(xpos+x) + qsin(ypos) >> 22) >> 21);
        memory = (ARGB *)((char *)memory + pitch);
        ypos ++;
    }
}


//=========================================================================//
// # 1
void inline draw_tile_block(ARGB * memory, long pitch, unsigned xpos, unsigned ypos)
{
    ypos = ypos * 3 - (Frame_step>>18);
    xpos = xpos * 3 + (Frame_step>>17);
    for( unsigned y = V_tile_size; y; y-- )
    {
        // 0xfe    ,    
        ARGB c = 0x806060 | 0xfe0000 & qsin(qsin(++ypos + ++xpos)>>24)>>8;
        for( unsigned x = 0; x < H_tile_size; x++)
            memory[x] = c | 0xfe00&(qsin(qsin(xpos+x) - qsin(ypos)>>24)>>16);
        memory[0] >>= 1;
        memory[H_tile_size-1] >>= 1;        
        memory = (ARGB *)((char *)memory + pitch);
    }
    //   2        .
    for( unsigned x = H_tile_size; x; )
    {
        *(ARGB *)((char *)memory + --x*sizeof(ARGB) - pitch) >>= 1;
        *(ARGB *)((char *)memory + x*sizeof(ARGB) - V_tile_size*pitch) >>= 1;
    }
}


//=========================================================================//
// # 2
void inline draw_tile_brick(ARGB * memory, long pitch)
{
    const   color = 0xCF2020;
    //    21   21  
    //   0xFFFFFFFE
    long    line = 1<<H_tile_size ^ ~(1<<3*7 | 077<<2*7 | 077<<1*7 | 077<<0*7)>>1;
    for( unsigned y = 2; y; y-- )
    {
        ARGB * p = memory;
        do
        {
            ARGB pixel = ((line & 1) - 1) & color;
            *(ARGB *)((char *)p + 13*pitch) =
            *(ARGB *)((char *)p + 12*pitch) =
            *(ARGB *)((char *)p + 11*pitch) =
            *(ARGB *)((char *)p + 10*pitch) =

            *(ARGB *)((char *)p + 03*pitch) =
            *(ARGB *)((char *)p + 02*pitch) =
            *(ARGB *)((char *)p + 01*pitch) =
            *p = pixel;

            *(ARGB *)((char *)p   + 14*pitch) = 
            *(ARGB *)((char *)p++ + 04*pitch) = 0;
            line >>= 1;
        }
        while( line != 0xFFFFFFFE );
        memory = (ARGB *)((char *)memory + 5*pitch);
        line = 1<<H_tile_size ^ ~(037<<3*7 | 077<<2*7 | 077<<1*7 | 077<<0*7)>>5;
    }
}


//=========================================================================//
// # 3
void draw_tile_bridge(ARGB * memory, long pitch)
{
    //  
    for( unsigned x = H_tile_size; x; )
    {
        *(ARGB *)((char *)memory + x*sizeof(ARGB) + 2*pitch) =
            memory[--x] = 0x888820;
        *(ARGB *)((char *)memory + x*sizeof(ARGB) + pitch) = 0xFFFF88;
    }
}


// ""  -   tile_bridge (#3)
void draw_tile_under_bridge(ARGB * memory, long pitch, unsigned ypos)
{
    const   pixels = 128;
    static ARGB c[pixels];
    for( unsigned y = V_tile_size; y; y-- )
    {
        int i = y + (Frame_step>>15) - ypos*V_tile_size;
        for( unsigned x = H_tile_size/2+1; x; )
        {
            i = ++i & pixels - 1;
            memory[--x+H_tile_size/2] += c[i];
        }
        for( unsigned x = H_tile_size/2; x; )
        {
            i = --i & pixels - 1;
            memory[--x] += c[i];
        }
        memory = (ARGB *)((char *)memory + pitch);
    }
    if( c[pixels/2-1] )
        return;

    //  ,   
    for( unsigned i = 0; i < pixels/2; i++ )
    {
        c[i] = 0x003F00 & sine[i*3*sine_table_size/pixels] >> 26-8
             | 0x3F0000 & sine[i*2*sine_table_size/pixels] >> 26-16 
               // sine[2*sine_table_size] - 2   !!!
               //     .
             | sine[i*4*sine_table_size/pixels] >> 26;           
    }
}


//=========================================================================//
// # 4
void inline draw_tile_treasure(ARGB * memory, long pitch, unsigned xpos, unsigned ypos)
{
    const phases = 1<<4;
    //    ,        
    static unsigned char sprite[phases][V_tile_size+1][H_tile_size];
    char * red   = (char *)&sprite[phases-1 & 8*ypos-(Frame_step>>20)];
    char * green = (char *)&sprite[phases-1 & 4*xpos+(Frame_step>>18)];
    char * blue  = (char *)&sprite[phases-1 & ((xpos^ypos)-(Frame_step>>19))];

    for( unsigned y = V_tile_size; y; y-- )
    {
        for( unsigned x = 0; x < H_tile_size; x++)
            memory[x] += *red++ | *green++ << 8 | *blue++ <<16 ;
        memory = (ARGB *)((char *)memory + pitch);
    }
    //    ,    
    if( sprite[0][V_tile_size/2][H_tile_size/2] )
        return;

    //  
    const R = V_tile_size/2;
    for( int i = phases - 1 ; i >= 0; i-- )
        for( int a = sine_table_size/2; a; a -= sine_table_size >> 5 )
            for( unsigned r = R + 1; r; r-- )
            {
                // 
                int x = r*(sine[a]>>16) >> 16;
                // 
                int y = r*(qsin(a+sine_table_size/2)>>16) >> 16;
                if( x + y <= R )
                {
                      sprite[i][R-x][R+y] = sprite[i][R+x][R-y]
                    = sprite[i][R-y][R-x] = sprite[i][R+y][R+x] 
                    = qsin(i*(sine_table_size/phases) + a) >> 25;
                }
            }
}


//=========================================================================//
// # 7
void inline draw_tile_ladder(ARGB * memory, long pitch)
{
    //  
    for( unsigned y = V_tile_size; y; y-- )
    {
        memory[1] = memory[3] = 
        memory[H_tile_size-4] = memory[H_tile_size-2] = 0x707070;
        memory[2] = memory[H_tile_size-3] = 0xA0A0A0;
        memory = (ARGB *)((char *)memory + pitch);

    }
    memory = (ARGB *)((char *)memory - 6 * pitch + 3*sizeof(ARGB));
    //     
    for( unsigned i = 2; i; i-- )
    {
        for( unsigned x = H_tile_size-6; x; )
        {
            *(ARGB *)((char *)memory + x*sizeof(ARGB) + 2*pitch) =
                memory[--x] = 0x707070;
            *(ARGB *)((char *)memory + x*sizeof(ARGB) + pitch) = 0xA0A0A0;
        }
        memory = (ARGB *)((char *)memory - V_tile_size/2 * pitch);
    }

}


//=========================================================================//
//   
void inline draw_room(
            ARGB * memory, 
            long pitch, 
            short (* level_data)[V_tiles * H_tiles / tiles_in_block] )
{
    ARGB * p = (ARGB *)((char *)memory + 
        + sizeof(ARGB)*(width-H_tiles*H_tile_size)/2 + vborder*pitch);

    cache_room_data(level_data);
    
    for( unsigned y = 0; y < V_tiles; y++ )
    {
        for( unsigned x = 0; x < H_tiles; x++ )
        {
            switch( room[y+1][x] )
            {
            case empty:
                draw_tile_empty(p+x*H_tile_size, pitch, x, y);
                break;
                
            case block:
                draw_tile_block(p+x*H_tile_size, pitch, x, y);
                break;

            case brick:
                draw_tile_brick(p+x*H_tile_size, pitch);
                break;

            case bridge | under_bridge:
                draw_tile_empty(p+x*H_tile_size, pitch, x, y);
                draw_tile_under_bridge(p+x*H_tile_size, pitch, y);
                draw_tile_bridge(p+x*H_tile_size, pitch);
                break;

            case treasure:
                draw_tile_empty(p+x*H_tile_size, pitch, x, y);
                draw_tile_treasure(p+x*H_tile_size, pitch, x, y);
                break;
                
            case treasure | under_bridge:
                draw_tile_empty(p+x*H_tile_size, pitch, x, y);
                draw_tile_under_bridge(p+x*H_tile_size, pitch, y);
                draw_tile_treasure(p+x*H_tile_size, pitch, x, y);
                break;
                
            case ladder:
                draw_tile_empty(p+x*H_tile_size, pitch, x, y);
                draw_tile_ladder(p+x*H_tile_size, pitch);
                break;

            case empty | under_bridge:
                draw_tile_empty(p+x*H_tile_size, pitch, x, y);
                draw_tile_under_bridge(p+x*H_tile_size, pitch, y);
                break;
            }
        }
        p = (ARGB *)((char *)p + V_tile_size*pitch);
    }
}


