/************************************************************************
 * 
 ************************************************************************
 */ 

#include <ddraw.h>
#include "pixel_format.h"


#define DRAW(X) stgfx::dispatcher(X);


namespace   stgfx
{
    
//=========================================================================//
// 
enum    { width = 640,  height = 480 };
enum    { sine_table_size = 1024 };
enum    { max_gfx = 255 };


//=========================================================================//
// 

static  void (__fastcall * gfx[max_gfx+1])(ARGB *, long);  //   = 0
static  int         initialized;


#include "gfx_sine_table.cpp"
#include "gfx_background_pattern.cpp"
#include "gfx_logo.cpp"
#include "gfx_text.cpp"
#include "gfx_menu.cpp"

void    select_root_menu();

#include "gfx_game.cpp"
#include "gfx_root.cpp"


//=========================================================================//


void static inline
dispatcher(IDirectDrawSurface7 * surface)
{
    TIMING_START

    if( ! initialized )
    {
        init_sine_table();
        init_background_pattern();       
        select_root_menu();
        --initialized;
    }
    DDSURFACEDESC2          descriptor;
    zero_memory(&descriptor, sizeof(descriptor));
    descriptor.dwSize |= sizeof(descriptor);
    if( surface->Lock(0, &descriptor, 0, 0) )
        panic("couldn't lock backbuffer surface");
    else
    {
        for( int i = 0; ; i++ )
        {
            if( (int)gfx[i] < 0 )
                continue;
            if( ! gfx[i] )
                break;
            gfx[i]((ARGB *)descriptor.lpSurface, descriptor.lPitch);
        }
        surface->Unlock(0);
    }

    TIMING_END("")
}


}

// EOF
