/************************************************************************
 * 
 ************************************************************************
 */ 

#pragma once
#include "cpu.cpp"


//=========================================================================//
#ifndef NO_TIMING
#define TIMING_START            \
    unsigned    __tsc;          \
{                               \
    __asm   rdtsc               \
    __asm   mov     __tsc, eax  \
}
#define TIMING_END(TEXT)            \
{                                   \
    __asm   rdtsc                   \
    __asm   sub     eax, __tsc      \
    __asm   xor     edx, edx        \
    __asm   div     cpu_frequency   \
    __asm   push    eax             \
}                                   \
    PRINTF(TEXT"%d\r\n");           \
{   __asm   pop     eax     }       
#else
#define TIMING_START
#define TIMING_END(FOO)
#endif


//=========================================================================//
#define SET_FPU_CONTROL_WORD(X) \
{                               \
    __asm   push    X           \
    __asm   fldcw   [esp]       \
    __asm   add     esp, 4      \
}
#define FPU_MASK_ALL_EXEPTIONS  0x7F
#define FPU_PRECISION_24        (0<<8)
#define FPU_PRECISION_53        (2<<8)
#define FPU_PRECISION_64        (3<<8)
#define FPU_ROUND_TO_NEAREST    (0<<10)
#define FPU_ROUND_DOWN          (1<<10)
#define FPU_ROUND_UP            (2<<10)
#define FPU_ROUND_TOWARD_ZERO   (3<<10)


//=========================================================================//
//       
//   ,       .
__declspec(naked) int __fastcall
safe_terminate_thread(void * thread)
{
    //       !!!
    static  CONTEXT     context;
    __asm
    {   // ecx = handle to thread.
        //  Esp      
        //   ,   TIB,
        //       .
        // ,     , 
        // .. esp    -   .
        //      
        //   esp   :
        // ,  Esp  ,      . 
        mov     context.ContextFlags, CONTEXT_CONTROL
        push    0                           //  TerminateThread
        push    ecx                         //  TerminateThread
        push    offset context
        push    ecx
        call    dword ptr GetThreadContext
        xor     edx, edx
        mov     eax, context.Esp
        cmp     dword ptr fs:[edx+4], eax   // Top of user stack
        setb    cl
        cmp     dword ptr fs:[edx+8], eax   // Base of user stack
        setae   ch
        test    cx, cx
        jz      current_thread
        call    dword ptr TerminateThread
        ret
current_thread:
        pop     ecx                         //   TerminateThread 
        pop     ecx
        ret
    }
}


//=========================================================================//
void inline __fastcall
zero_memory(void * destination, unsigned count)
{
fill:
    *(char *)destination = 0;
    destination = (char *)destination + 1;
    if(--count)
        goto fill;
}


//=========================================================================//
void inline 
zero_stack()
{
    __asm
    {
        xor     ecx, ecx
        inc     ch
        mov     eax, esp
repeat: push    0
        loop    repeat
        xchg    esp, eax
    }
}


//=========================================================================//
void inline
copy_by_dwords(void * destination, void * source, unsigned bytes)
{
    __asm   mov     ecx, bytes
    __asm   mov     esi, source
    __asm   mov     edi, destination
    __asm   shr     ecx, 2
    __asm   rep     movsd
}


//=========================================================================//
unsigned inline qrandom()
{
    __asm   rdtsc
    __asm   mul     edx
    __asm   xor     eax, edx
}


// EOF
