/************************************************************************
 * 
 ************************************************************************
 */ 

#pragma once

#include "log_window.cpp"
#include "keyboard.cpp"

#include "pixel_format.h"

#include <ddraw.h>

#ifndef DRAW
#define DRAW(X)
#endif


class   user_interface :    public  window, input_device
{

public:

inline
user_interface(const char * name, const int width, const int height)
    : window()
{
    //    !
    this->name          = name;
    this->width        |= width;
    this->height       |= height;
    this->frame        |= backbuffer1;
  //  this->fullscreen   |= ~0;
    InitializeCriticalSection(&busy);
}

void kill();


void inline
start_engine(const int priority = THREAD_PRIORITY_HIGHEST)
{
    unsigned long dummy_ID;
    thread = CreateThread(0, 0, show_frames, this, 0, &dummy_ID);
    SetThreadPriority(thread, priority);
}

protected:

enum   surfases {primary, backbuffer1, backbuffer2, quantity_of_surfaces};

    IDirectDraw7          * direct_draw;
    IDirectDrawSurface7   * surface[quantity_of_surfaces];
    IDirectDrawClipper    * clipper;

    unsigned                max_visible_scanlines;
    int                     frame;                     
    unsigned                refresh_rate;
    int                     fullscreen;
    int                     minimized;
    void                  * thread;
    CRITICAL_SECTION        busy;

#define WM_USER_SWITCH_SCREEN_MODE WM_USER

void create_direct_draw();

void create_surfases();

void release_surfases();


static unsigned long __stdcall
show_frames(void * p)
{
    user_interface * This = (user_interface *)p;

    while( true )
    {
        //      (),  .
        while( true )
        {
            Sleep(1);
            if( This->clipper )
            {
                unsigned long   size;
                This->clipper->GetClipList(0, 0, &size);
                if( size < sizeof(RGNDATA) )
                    Sleep(50);
                else
                    break;
            }
            else if( ! IsIconic(This->handle) )
                break;
        }
        This->process_system_keys();
        EnterCriticalSection(&This->busy);

        //     .
        while( true )
        {
            unsigned long   scanline;
            if( This->direct_draw->GetScanLine(&scanline) == DDERR_UNSUPPORTED )
            {
                //    CPU
                //      GetScanLine
                This->direct_draw->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, 0);
                break;
            }
            if( scanline >= This->max_visible_scanlines - 50)
                break;
            Sleep(This->max_visible_scanlines - scanline
                > This->max_visible_scanlines * This->refresh_rate / 800 );
        }

        switch( This->direct_draw->TestCooperativeLevel() )
        {
        case DDERR_WRONGMODE:
            if( ! This->fullscreen )
            {
                This->release_surfases();
                This->direct_draw->Release();
                This->create_direct_draw();
                This->create_surfases();
            }
        case DD_OK:
            RECT        destination;
            destination.left = destination.top = 0;
            ClientToScreen(This->handle, (POINT *)&destination);
            destination.right = destination.left + This->width;
            destination.bottom = destination.top + This->height;
            if( This->surface[primary]->IsLost() )
                This->direct_draw->RestoreAllSurfaces();
            if( This->surface[primary]->Blt(&destination, 
                    This->surface[This->frame],  0, DDBLT_DONOTWAIT, 0) )
                This->warn_unsupported_mode();
            else
            {
                Frame_step += 60*0x10000 
                    /(This->refresh_rate ? This->refresh_rate : 60);

                This->frame ^= backbuffer1 | backbuffer2;
                DRAW(This->surface[This->frame]);

                for( int key = DIK_0; key >= DIK_1; key-- )
                    if( This->is_key_hit(key) )
                        cps_player::choose(key - DIK_1 + 2);
            }
        }
        LeaveCriticalSection(&This->busy);
    }// while( true )
    return 0;
}


virtual void
window_procedure(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch( msg )
    {
    case WM_ACTIVATE:
        int is_iconic;  is_iconic = HIWORD(wParam);
        if( fullscreen )
            if( LOWORD(wParam) == WA_INACTIVE )
            {
                if( ! minimized )
                    ShowCursor(1);
            }
            else
            {
                if( ! is_iconic )
                    ShowCursor(0);
            }
        if( minimized != is_iconic )
            for( int n = quantity_of_surfaces-1; n > primary; n-- )
            {
                if( ! is_iconic && surface[n]->PageLock(0) )
                    panic("couldn't lock backbuffer's pages");
                if( is_iconic && surface[n]->PageUnlock(0) )
                    panic("couldn't unlock backbuffer's pages");
            }
        minimized = is_iconic;
        break;

    case WM_MOVE:
        if( ! fullscreen )
        {
            RECT    rectangle;
            GetWindowRect(handle, &rectangle);
            x = rectangle.left; y = rectangle.top;
        }
        break;

    case WM_CREATE:
        input_device::initialize(handle);
        create_direct_draw();
        create_surfases();
        break;

    // ""      .
    case WM_WINDOWPOSCHANGING:
        if( ((WINDOWPOS *)lParam)->flags & SWP_SHOWWINDOW )
        {
            RECT    client;
            GetClientRect(handle, &client);
            ((WINDOWPOS *)lParam)->cx = 2 * width - client.right;
            ((WINDOWPOS *)lParam)->cy = 2 * height - client.bottom;
            ((WINDOWPOS *)lParam)->flags &= ~SWP_NOSIZE;
        }
        break;

    case WM_USER_SWITCH_SCREEN_MODE:
        EnterCriticalSection(&busy);
        ShowCursor(fullscreen);
        release_surfases();
        fullscreen = ~fullscreen;
        create_surfases();
        LeaveCriticalSection(&busy);
        break;

    case WM_DESTROY:
        PostQuitMessage(0);
    }
}


void inline
process_system_keys()
{
    input_device::get_state();
    if( is_key_pressed(DIK_LMENU)  || is_key_pressed(DIK_RMENU) )
    {
        if( is_key_hit(DIK_RETURN) || is_key_hit(DIK_NUMPADENTER))
            PostMessageA(handle, WM_USER_SWITCH_SCREEN_MODE, 0, 0);
        else if( is_key_pressed(DIK_F4) )
            PostMessageA(handle, WM_CLOSE, 0, 0);
    }
}


void inline
warn_unsupported_mode()
{
    Sleep(50);
    HDC device_context = GetDC(handle);
    BitBlt(device_context, 0, 0, width, height, 0, 0, 0, WHITENESS);
    SetBkColor(device_context, 0xFFFFFF);
    SetTextColor(device_context, 0x7F);
    TextOutA(device_context, 32, height / 3,
        "Only 32 BPP windowed mode is supported. "
        "Press Alt+Enter to switch to full-screen mode.", 86);
}

};

//=========================================================================//


void user_interface::kill()
{
    safe_terminate_thread(thread);
 //   DeleteCriticalSection(&busy);
    release_surfases();
    if( clipper )
        clipper->Release();
    if( direct_draw )
        direct_draw->Release();
    clipper     = 0;
    direct_draw = 0;
    input_device::kill();
}

void user_interface::release_surfases()
{
    if( direct_draw )
    {
        direct_draw->RestoreDisplayMode();
        for(int n = quantity_of_surfaces; n;)
        {
            if( surface[--n] )
                surface[n]->Release();
            surface[n] = 0;
        }
    }
}

void user_interface::create_direct_draw()
{
    PRINTF("initializing DirectDraw...\r\n");
    int(__stdcall * DirectDrawCreateEx)(GUID*, void**, REFIID, void*)
        = (int(__stdcall *)(GUID*, void**, REFIID, void*))
        GetProcAddress(LoadLibraryA("ddraw"),"DirectDrawCreateEx");
    //  DirectDraw ,  ""  UnhandledExceptionFilter
    if( ! DirectDrawCreateEx )
        panic("DirectX 7 or later not installed");
    if( DirectDrawCreateEx(0, (void**)&direct_draw,
                            IID_IDirectDraw7, 0) )
        panic("couldn't initialize DirectDraw");
}

void user_interface::create_surfases()
{
    DDSURFACEDESC2          surface_descriptor;

    zero_memory(&surface_descriptor, sizeof(surface_descriptor));
    // OR  3 :
    surface_descriptor.dwSize |= sizeof(surface_descriptor);
    
    PRINTF("switching %s mode\r\n",
                fullscreen ? "fullscreen" : "windowed");
    
    SetWindowLongA(handle, GWL_STYLE,    // fullscreen -1  0
                fullscreen &(WS_POPUP|WS_SYSMENU) |
                // WS_VISIBLE  
                //  WM_WINDOWPOSCHANGING   .
                //    ShowWindow .
               ~fullscreen &(WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX) );

    if( direct_draw->SetCooperativeLevel(handle,
                fullscreen &(DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN) |
               ~fullscreen & DDSCL_NORMAL ) )
        panic("couldn't set proper cooperative level");

    if( ! fullscreen )
    {
        SetWindowPos(handle, 0, x, y, width, height, 
                    SWP_FRAMECHANGED|SWP_NOZORDER);
        ShowWindow(handle, SW_SHOW);
        //       .
        // (-  XP   RestoreDisplayMode ?!)
        direct_draw->SetDisplayMode(0, 0, 0, 0, 0);
        if( ! clipper
           && direct_draw->CreateClipper(0, &clipper, 0) )
            panic("couldn't create clipper");
        else
            clipper->SetHWnd(0, handle);
    }
    else if( direct_draw->SetDisplayMode(width, height, BPP, 0, 0) )
        panic("couldn't change display mode");

    surface_descriptor.dwFlags        |= DDSD_CAPS;
    surface_descriptor.ddsCaps.dwCaps |= DDSCAPS_PRIMARYSURFACE;

    if( direct_draw->CreateSurface(&surface_descriptor, &surface[primary], 0) )
        panic("couldn't create primary surface");

    if( ! fullscreen )
    {
        surface[primary]->SetClipper(clipper);
    }
    surface_descriptor.dwFlags |= DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS | DDSD_PIXELFORMAT ;
    surface_descriptor.dwWidth  = width;
    surface_descriptor.dwHeight = height;
    surface_descriptor.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN |
                                        DDSCAPS_SYSTEMMEMORY;
                                     //   DDSCAPS_NONLOCALVIDMEM|DDSCAPS_VIDEOMEMORY;
    surface_descriptor.ddpfPixelFormat.dwSize |= sizeof(DDPIXELFORMAT);
    surface_descriptor.ddpfPixelFormat.dwFlags |= DDPF_RGB;
    surface_descriptor.ddpfPixelFormat.dwRGBBitCount |= BPP;
    surface_descriptor.ddpfPixelFormat.dwRBitMask |= 0xFF << RED_SHIFT;
    surface_descriptor.ddpfPixelFormat.dwGBitMask |= 0xFF << GREEN_SHIFT;
    surface_descriptor.ddpfPixelFormat.dwBBitMask |= 0xFF << BLUE_SHIFT;
    for( int n = quantity_of_surfaces-1; n > primary; n-- )
    {
        if( direct_draw->CreateSurface(&surface_descriptor, &surface[n], 0))
            panic("couldn't create backbuffer surface");
        if( surface[n]->PageLock(0) )
            panic("couldn't lock backbuffer's pages");
    }    
    direct_draw->GetDisplayMode(&surface_descriptor);
    PRINTF("Ok: %dx%d@%dHz\r\n",
                                surface_descriptor.dwWidth,
        max_visible_scanlines = surface_descriptor.dwHeight,
        refresh_rate          = surface_descriptor.dwRefreshRate);
}


// EOF
